// spxlSubAdd Effect Plugin
// More or less the same result as the AddSubtract sample plugin, with extra parameters for testing purposes.
// The "multiplier" parameter is used to scale the R,G,B addition/subtraction.

// 2019-08-12 ffgl@subpixels.com
// - Created, based on AddSubbtract sample code

#include "spxlSubAdd.h"
#include "ffgl/ParamBool.h"
#include "ffgl/ParamBuffer.h"
#include "ffgl/ParamFloat.h"
#include "ffgl/ParamInteger.h"
#include "ffgl/ParamOption.h"
#include "ffgl/ParamRange.h"
#include "ffgl/ParamText.h"
#include "ffglquickstart/ParamEvent.h"
#include "ffglquickstart/ParamFFT.h"
#include "ffglquickstart/ParamTrigger.h"

using namespace ffgl;
using namespace ffglex;
using namespace ffglqs;

namespace spxl
{
namespace ffgl
{
namespace effect
{
static ::ffgl::PluginInfo pluginInfo(
	PluginFactory< SubAdd >,// Create method
#ifdef _DEBUG
	"?e01",   // Plugin unique ID of maximum length 4.
	"?SubAdd",// Plugin name
#else
	"_e01",   // Plugin unique ID of maximum length 4.
	"_SubAdd",// Plugin name
#endif
	2,                    // API major version number
	1,                    // API minor version number
	1,                    // Plugin major version number
	0,                    // Plugin minor version number
	PluginType::FF_EFFECT,// Plugin type

	"Subtract and Add colours",                        // Plugin description
	"Resolume FFGL Example (mod by ffgl@subpixels.com)"// About
);

static const char _fragmentShaderCode[] = R"(
void main()
{
	vec4 color = texture( inputTexture, i_uv );
	//The InputTexture contains premultiplied colors, so we need to unpremultiply first to apply our effect on straight colors.
	if( color.a > 0.0 )
		color.rgb /= color.a;

	color.rgb += adjustment * scaleAdj;

	if (doAdd) color.rgb += colAdd.rgb * scaleAdd;
	if (doSub) color.rgb -= colSub.rgb * scaleSub;

	//The plugin has to output premultiplied colors, this is how we're premultiplying our straight color while also
	//ensuring we aren't going out of the LDR the video engine is working in.
	color.rgb = clamp( color.rgb * color.a, vec3( 0.0 ), vec3( color.a ) );
	fragColor = color;
}
)";

SubAdd::SubAdd()
{
	// We declare that this plugin has a Brightness parameter which is a RGB param.
	// The uniform name ("adjustment") must match the uniform name in the fragment shader.
	adjustment = AddRGBRangeParam( "Adjust", "adjustment", 0.0f, { -1.0f, 1.0f } );

	AddParam( new ParamFloat( "Scale", "scaleAdj", 1.0f ) );

	col1 = AddHueColorParam( "Add", "colAdd", HSBAColor::BLACK );
	col2 = AddHueColorParam( "Subtract", "colSub", HSBAColor::BLACK );

	AddParam( new ParamFloat( "Add Scale", "scaleAdd", 1.0f ) );
	AddParam( new ParamFloat( "Sub Scale", "scaleSub", 1.0f ) );

	AddParam( new ParamBool( "Do Add", "doAdd", true ) );
	AddParam( new ParamBool( "Do Sub", "doSub", true ) );

	// Test other parameter types

	AddParam( new ParamFloat( "Test X pos", ParameterType::FF_TYPE_XPOS, "testX" ) );
	AddParam( new ParamFloat( "Test Y pos", ParameterType::FF_TYPE_YPOS, "testY" ) );

	AddParam( new ParamFloat( "TestBrightness", ParameterType::FF_TYPE_BRIGHTNESS ) );
	AddParam( new ParamFloat( "TestSaturation", ParameterType::FF_TYPE_SATURATION ) );
	AddParam( new ParamFloat( "TestAlpha", ParameterType::FF_TYPE_ALPHA ) );
	AddParam( new ParamFloat( "TestHue", ParameterType::FF_TYPE_HUE ) );

	AddParam( new ParamEvent( "TestEvent" ) );
	AddAudioParam( new ParamFFT( "TestFFT", 8 ) );

	testOption = new ParamOption( "TestOption", 2U );
	testOption->AddOption( { "Opt 1", 11.1f } );
	testOption->AddOption( { "Opt 2", 22.2f } );
	testOption->AddOption( { "Opt 3", 33.3f } );
	testOption->AddOption( { "Opt 4", 44.4f } );
	testOption->AddOption( { "Opt 5", 55.5f } );
	AddParam( testOption );

	AddParam( new ParamRange( "TestRange", "temperature", 25.0f, { 0.0f, 100.0f } ) );
	AddParam( new ParamInteger( "TestInteger", "xscale", 0, { -1920.0f, 1920.0f } ) );
	AddParam( new ParamText( "TestText", "Lorem ipsum blah blah amet" ) );
	AddParam( new ParamTrigger( "TestTrigger" ) );

	SetFragmentShader( _fragmentShaderCode );

	SetTimeSupported( true );
}

SubAdd::~SubAdd()
{
}

void SubAdd::Update()
{
	// Try out accessing parameter values

	float r = adjustment.red->GetFloatValue();
	float g = adjustment.green->GetFloatValue();
	float b = adjustment.blue->GetFloatValue();

	float average = ( r + g + b ) / 3.0f;

	RGBAColor rgba = adjustment.GetColor();
	float red      = rgba.r;

	HSBAColor color1 = col1.GetColor();
	HSBAColor color2 = col2.GetColor();

	float hue1 = color1.hue;

	std::string uniformName = col1.hue->GetUniformName();

	FFUInt32 selectedOption   = testOption->GetIndex();
	std::string selectedName  = testOption->GetOptionName( selectedOption );

	float selectedValue;
	FFResult res = testOption->GetOptionValue( selectedOption, selectedValue );

	float selectedValueDirect = testOption->GetFloatValue();
}

}// namespace effect
}// namespace ffgl
}// namespace spxl
